/* set-user-1.1--1.4.sql */

-- complain if script is sourced in psql, rather than via ALTER EXTENSION
\echo Use "ALTER EXTENSION set_user UPDATE to '1.4'" to load this file. \quit

-- just bumping our version to 1.4. no new features here, so nothing to do.
